-------------------------------
--  Advanced Shadows (ASHA)  --
--      For X-Plane 11       --
--   Developed by Alreiber   --
-------------------------------
local asha_version = "2.2.1"
local asha_build = "2020.08.31"
-------------------------------

-- Startup timer
local asha_startup_clock = os.clock() + 1
local asha_started = false
-- Main container of ASHA
function asha_main_loop()
	-- SKIP EXECUTION IF DELAY NOT FINISHED OR THIS FUNCTION HAS ALREADY STARTED
	if os.clock() < asha_startup_clock or asha_started then
		return
	end
	asha_started = true
	
	-- FUNCTION: FPS COUNTER
	local start_time = os.clock()
	local frame_count = 0
	local elapsed_time = 0
	local fps = 0
	local PERIOD = 0.5 -- seconds
	function count_frames()
		elapsed_time = os.clock() - start_time
		if elapsed_time < PERIOD then
			frame_count = frame_count + 1
		else
			start_time = os.clock()
			fps = frame_count / elapsed_time
			frame_count = 1
		end
	end
	
	-- FUNCTION: FILE EXISTS
	function exists(file)
	   local ok, err, code = os.rename(file, file)
	   if not ok then
		  if code == 13 then
			 -- Permission denied, but it exists
			 return true
		  end
	   end
	   return ok
	end

	--- FUNCTION: DIR EXISTS
	function isdir(path)
	   return exists(path .. "/")
	end
	
	-- READ CURRENT X-PLANE VERSION
	local is1150 = tonumber(string.sub(XPLANE_VERSION, 1, 4)) >= 1150 -- flag to tell if this is a post-11.50 X-Plane

	--FUNCTION: READ X-PLANE.PRF
	-- Reads a key in X-Plane.prf
	function xpprf_read(key)
		-- open X-Plane.prf (try with the temp file first if it exists)
		local file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf.tmp", "r")
		if file==nil then
			file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf", "r")
		end
		-- read file
		local content = file:read("*a")
		file:close()
		-- return the value
		local value = string.match(content, key .. "%s+([^\n]+)")
		if tonumber(value) ~= nil then
			value = tonumber(value)
		end
		return value
	end
	
	-- FUNCTION: WRITE X-PLANE.PRF
	-- Writes a value in a key in X-Plane.prf. As X-Plane overwrites the existing X-Plane.prf on close,
	-- we do the changes in a temporary file and create a batch file that is executed after X-Plane closes.
	-- This batch file will replace the X-Plane.prf that X-Plane writes on close with the temporary file.
	function launch_batch()
		--os.execute("wscript.exe \"" .. SYSTEM_DIRECTORY .. "Output/preferences/asha.vbs\" \"" .. SYSTEM_DIRECTORY .. "Output/preferences/asha.bat\"")
		os.execute("start wscript.exe \"" .. SYSTEM_DIRECTORY .. "Output\\preferences\\asha.vbs\" \"" .. SYSTEM_DIRECTORY .. "Output\\preferences\\asha.bat\"")
	end
	function xpprf_write(key, value)
		-- check this key value first, if it is identical don't do anything
		if xpprf_read(key)==value then
			return
		end
		-- open X-Plane.prf.tmp
		local file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf.tmp", "r")
		if file==nil then
			file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf", "r")
		end
		-- read file
		local content = file:read("*a")
		file:close()
		-- replace the key value
		content = string.gsub(content, key .. "%s+[^\n]+", key .. " " .. tostring(value))
		-- write new contents to file
		file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf.tmp", "w")
		file:write(content)
		file:close()
		-- stop here if batch file already exists
		file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/asha.bat", "r")
		if file ~= nil then
			return
		end
		-- create batch file
		file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/asha.bat", "w")
		content = "ping localhost -n 5 >NUL\ndel \"" .. SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf\"\nrename \"" .. SYSTEM_DIRECTORY .. "Output/preferences/X-Plane.prf.tmp\" X-Plane.prf\ndel \"" .. SYSTEM_DIRECTORY .. "Output/preferences/asha.vbs\"\ndel \"" .. SYSTEM_DIRECTORY .. "Output/preferences/asha.bat\""
		content = string.gsub(content, "/", "\\")
		file:write(content)
		file:close()
		file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/asha.vbs", "w")
		file:write("Set WshShell = WScript.CreateObject(\"WScript.Shell\")\nWshShell.Run \"\"\"\" & WScript.Arguments(0) & \"\"\"\" & sargs, 0, False")
		file:close()
		-- execute batch file on Lua exit
		do_on_exit("launch_batch()")
	end

	-- GET CURRENT GRAPHIC SETTINGS
	-- read current "effects" settings from X-Plane.prf file
	local renopt_effects_04 = xpprf_read("renopt_effects_04")
	local renopt_scenery_shadows = xpprf_read("renopt_scenery_shadows")
	-- number of splits in cascaded shadow mapping
	local csm_split_exterior = get("sim/private/controls/shadow/csm_split_exterior")
	
	-- FUNCTION: READ AND WRITE SETTINGS
	function read_asha_settings_all()
		local file = io.open(SCRIPT_DIRECTORY .. "asha.ini", "r")
		if file==nil then
			return nil
		end
		local content = file:read("*a")
		file:close()
		local matches = string.gmatch(content, "[^\n%s]+%s*=%s*[^\n%s]+")
		local settings = {}
		local key
		local value
		for this_line in matches do
			key = string.match(this_line, "([^%s]+)%s*=")
			value = string.match(this_line, "=%s*([^%s]+)")
			if key ~= nil and value ~= nil then
				if value == "true" then
					settings[key] = true
				elseif value == "false" then
					settings[key] = false
				elseif tonumber(value) ~= nil then
					settings[key] = tonumber(value)
				else
					settings[key] = value
				end
			end
		end
		return settings
	end
	function read_asha_settings(key)
		local settings = read_asha_settings_all()
		local success, value = pcall(function(x, y) return x[y] end, settings, key)
		if success then
			return value
		else
			return nil
		end
	end
	function write_asha_settings(key, value)
		local settings = read_asha_settings_all()
		if settings == nil then
			settings = {}
		end
		settings[key] = value
		local content = "ASHA CONFIGURATION FILE"
		for a_key, a_value in pairs(settings) do
			content = content .. "\n" .. a_key .. " = " .. tostring(a_value)
		end
		local file = io.open(SCRIPT_DIRECTORY .. "asha.ini", "w")
		file:write(content)
		file:close()
	end
	
	-- READ SHADOW RESOLUTION SETTINGS IN settings.txt
	local RES_IN_SETTINGS = -1
	function res_in_settings_txt()
		-- read settings.txt
		local file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "r")
		local content = file:read("*a")
		file:close()
		-- get set of resolutions
		local res_min = 999999
		local res_max = -999999
		for res in string.gmatch(content, "fbo/shadow_cam_size%s+(%d+)") do
			res_min = math.min(res_min, res)
			res_max = math.max(res_max, res)
		end
		-- return current settings based on resolutions in settings.txt and cache it in RES_IN_SETTINGS
		if res_min==res_max then
			RES_IN_SETTINGS = res_min
			return res_min
		else
			RES_IN_SETTINGS = -1
			return -1
		end
	end
	-- read it at least once
	res_in_settings_txt()
	
	-- READ ASHA PREFERENCES
	-- debug mode (if true, overlay data used for shadows correction and corrections applied)
	local DEBUG_MODE = read_asha_settings("debug")
	if DEBUG_MODE == nil then
		DEBUG_MODE = false
		write_asha_settings("debug", false)
	end
	-- first run flag (if true, pop up a welcome message)
	local FIRST_RUN = read_asha_settings("first_run")
	if FIRST_RUN == nil or FIRST_RUN == true then
		FIRST_RUN = true
		write_asha_settings("first_run", false)
	end
	-- shadows switch (turns shadows correction on and off)
	local SHADOWS_ON = read_asha_settings("shadows")
	if SHADOWS_ON == nil then
		SHADOWS_ON = true
		write_asha_settings("shadows", true)
	end
	-- jet heat switch (turns jet heat correction on and off)
	local HEAT_ON = read_asha_settings("heat")
	if HEAT_ON == nil then
		HEAT_ON = true
		write_asha_settings("heat", true)
	end
	-- water fft switch (turns water fft correction on and off)
	local WATER_ON = read_asha_settings("water")
	if WATER_ON == nil then
		if is1150 then
			WATER_ON = true
			write_asha_settings("water", true)
		else
			WATER_ON = false
			write_asha_settings("water", false)
		end
	end
	-- aurora borealis switch (turns aurora borealis on and off)
	local AURORA_ON = read_asha_settings("aurora")
	if AURORA_ON == nil then
		AURORA_ON = true
		write_asha_settings("aurora", true)
	end
	-- force settings in ASHA preferences if it differs X-Plane preferences
	local xpprf_aurora_on = xpprf_read("renopt_aurora")==1
	if AURORA_ON ~= xpprf_aurora then
		if AURORA_ON then
			xpprf_write("renopt_aurora", 1)
		else
			xpprf_write("renopt_aurora", 0)
		end
	end
	-- fires and ballons switch (turns fires and balloons on and off)
	local FIRES_ON = read_asha_settings("fires")
	if FIRES_ON == nil then
		FIRES_ON = false
		write_asha_settings("fires", false)
	end
	-- force settings in ASHA preferences if it differs X-Plane preferences
	local xpprf_fires_on = xpprf_read("renopt_fire_ball")==1
	if FIRES_ON ~= xpprf_fires_on then
		if FIRES_ON then
			xpprf_write("renopt_fire_ball", 1)
		else
			xpprf_write("renopt_fire_ball", 0)
		end
	end
	-- shadows drawing distance
	local DRAW_DIST = read_asha_settings("draw_dist")
	if DRAW_DIST == nil then
		DRAW_DIST = 3000
		write_asha_settings("draw_dist", DRAW_DIST)
	end
	-- target FPS
	local TARGET_FPS = read_asha_settings("target_fps")
	if TARGET_FPS == nil then
		TARGET_FPS = 30
		write_asha_settings("target_fps", TARGET_FPS)
	end
	-- shadow resolution selected in ASHA menu (-1 is X-Plane defaults)
	local SHADOW_RES = read_asha_settings("shadow_res")
	if SHADOW_RES == nil then
		SHADOW_RES = res_in_settings_txt()
		write_asha_settings("shadow_res", SHADOW_RES)
	end
	-- number of shadow map splits
	local CSM_SPLITS_EXT = read_asha_settings("splits_ext")
	if CSM_SPLITS_EXT == nil then
		CSM_SPLITS_EXT = 3
		write_asha_settings("splits_ext", CSM_SPLITS_EXT)
	end
	local CSM_SPLITS_INT = read_asha_settings("splits_int")
	if CSM_SPLITS_INT == nil then
		CSM_SPLITS_INT = 3
		write_asha_settings("splits_int", CSM_SPLITS_INT)
	end
	-- fps meter
	local FPS_METER = read_asha_settings("fps_meter")
	if FPS_METER == nil then
		FPS_METER = false
		write_asha_settings("fps_meter", FPS_METER)
	end
	-- walk mode
	local WALK_MODE = read_asha_settings("walk_mode")
	if WALK_MODE == nil then
		WALK_MODE = false
		write_asha_settings("walk_mode", WALK_MODE)
	end
	-- latest known ASHA version
	local asha_version_previous = read_asha_settings("version")
	if asha_version_previous == nil then
		asha_version_previous = asha_version
	end
	write_asha_settings("version", asha_version)
	
	-- BIND CURRENT SHADOWS RESOLUTION
	DataRef("current_res", "sim/private/controls/fbo/shadow_cam_size", "readonly")
	local selected_res = current_res -- shadows resolution currently selected (the currently active resolution by default)
	local res_change_pending = 0 -- flag to indicate if a shadows resolution change is pending to be applied in the settings dialog. 0: no changes pending, 1: apply pending, 2: X-Plane restart pending
	-- ADDITIONAL VARIABLES TO BIND
	DataRef("camera_x", "sim/graphics/view/view_x", "readonly") -- active camera X-axis (north/south) position in global coordinates (meters)
	DataRef("camera_y", "sim/graphics/view/view_y", "readonly") -- active camera Y-axis (up/down) position in global coordinates (meters)
	DataRef("camera_z", "sim/graphics/view/view_z", "readonly") -- active camera Z-axis (east/west) position in global coordinates (meters)
	DataRef("altitude_agl", "sim/flightmodel/position/y_agl", "readonly") -- aircraft height above ground level (meters)
	DataRef("local_x", "sim/flightmodel/position/local_x", "readonly") -- aircraft default CG X-axis (north/south) position in global coordinates (meters)
	DataRef("local_y", "sim/flightmodel/position/local_y", "readonly") -- aircraft default CG Y-axis (up/down) position in global coordinates (meters)
	DataRef("local_z", "sim/flightmodel/position/local_z", "readonly") -- aircraft default CG Z-axis (east/west) position in global coordinates (meters)
	DataRef("is_external_view", "sim/graphics/view/view_is_external", "readonly") -- flag to indicate if active view is external to the aircraft
	DataRef("cam_pitch", "sim/graphics/view/view_pitch", "readonly") -- active camera pitch angle (0 deg: side view, 90 deg: top view)
	DataRef("vert_fov", "sim/graphics/view/vertical_field_of_view_deg", "readonly") -- vertical field of view (degrees)
	DataRef("horz_fov", "sim/graphics/view/field_of_view_deg", "readonly") -- horizontal field of view (degrees)
	DataRef("sun_pitch", "sim/graphics/scenery/sun_pitch_degrees", "readonly") -- sun pitch angle (degrees)
	DataRef("int_camera_heading", "sim/graphics/view/pilots_head_psi", "readonly") -- interior camera heading (degrees)
	DataRef("int_camera_z", "sim/graphics/view/pilots_head_z", "readonly") -- interior camera Z-axis (front/rear) position relative to CG (meters)
	DataRef("int_camera_x", "sim/graphics/view/pilots_head_x", "readonly") -- interior camera X-axis (left/right) position relative to CG (meters)
	DataRef("gs", "sim/flightmodel/position/groundspeed", "readonly") -- ground speed (meters/second)
	DataRef("parkbrake", "sim/flightmodel/controls/parkbrake", "readonly") -- parking brake
	-- variables bind only for debug mode
	if DEBUG_MODE then
		DataRef("cockpit_near_adjust", "sim/private/controls/shadow/cockpit_near_adjust", "readonly") -- span of the "near" shadow map
		DataRef("csm_lambda", "sim/private/controls/shadow/csm/lambda", "readonly") -- cascaded shadow maps lambda value
		if is1150 then
			DataRef("far_limit_exterior", "sim/private/controls/shadow/csm/far_limit_exterior", "readonly") -- span of the "far" shadow map in exterior views
			DataRef("far_limit_interior", "sim/private/controls/shadow/csm/far_limit_interior", "readonly") -- span of the "far" shadow map in interior views
		else
			DataRef("far_limit", "sim/private/controls/shadow/csm/far_limit", "readonly") -- span of the "far" shadow map in interior views
		end
	end

	-- FUNCTION: APPLY SHADOWS RESOLUTION CHANGE
	-- The resolution change has to be done in the settings.txt file as its associated DataRef is read only. A backup of default X-Plane file is done prior to any change
	local function set_shadow_resolution(resolution)
		-- if resolution is -1, RESTORE DEFAULT X-PLANE settings file
		if resolution == -1 then
			local file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt.bak", "r")
			if file~=nil then
				local content = file:read("*a")
				file:close()
				file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "w")
				file:write(content)
				file:close()
			end
		-- if resolution is different from -1, APPLY NEW RESOLUTION to settings.txt
		else
			-- read settings.txt
			local file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "r")
			local content = file:read("*a")
			file:close()
			-- create a backup if it does not exist
			file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt.bak", "r")
			if file~=nil then
				file:close()
			else
				file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt.bak", "w")
				file:write(content)
				file:close()
			end
			-- set new resolution in settings.txt
			content = string.gsub(content, "fbo/shadow_cam_size%s+%d+", string.format("fbo/shadow_cam_size\t%.0f", resolution))
			-- save changes to settings.txt
			file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "w")
			file:write(content)
			file:close()
		end
	end
	
	-- SET NUMBER OF CASCADED SHADOW MAPS SPLITS
	function set_csmsplits()
		-- read settings.txt
		local file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "r")
		local content = file:read("*a")
		file:close()
		-- create a backup if it does not exist
		file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt.bak", "r")
		if file~=nil then
			file:close()
		else
			file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt.bak", "w")
			file:write(content)
			file:close()
		end
		-- set new resolution in settings.txt
		content = string.gsub(content, "shadow/csm_split_exterior%s+%d+", "shadow/csm_split_exterior " .. tostring(CSM_SPLITS_EXT))
		content = string.gsub(content, "shadow/csm_split_interior%s+%d+", "shadow/csm_split_interior " .. tostring(CSM_SPLITS_INT))
		-- save changes to settings.txt
		file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "w")
		file:write(content)
		file:close()
	end

	-- START UP WINDOW
	-- it displays several stuff that must be notified to the user at start up (welcome at first run, settings change, etc)
	-- FIRST, THE DRAWING FUNCTION
	-- we have to declare all variables used in the function, though they are actually created AFTER the function definition,
	-- otherwise the function would use a different copy of those variables
	local welcome_window = false
	local newrelease_window = false
	local shadowres_window = false
	local csmsplit_window = false
	local sceneryshadows_window = false
	local shadowresmismatch_window = false
	local reboot_required = false
	local startup_window
	-- draw window contents
	function draw_startup_window(wnd, x, y)
		if welcome_window then
			imgui.TextUnformatted("Welcome to ASHA " .. asha_version .. "!")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("New settings menu")
			imgui.TextUnformatted("-----------------")
			imgui.TextUnformatted("Check out the new ASHA settings menu at:")
			imgui.TextUnformatted("  Plugins -> FlyWithLua -> FlyWithLua Macros -> ASHA settings")
			imgui.TextUnformatted("You will find tweaks for shadows, heat effect, water simulation...")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Shadows advice")
			imgui.TextUnformatted("--------------")
			imgui.TextUnformatted("ASHA initially keeps your default X-Plane shadows resolution, in order to boost")
			imgui.TextUnformatted("performance, as ASHA shadow optimizations are independent of shadow resolution.")
			imgui.TextUnformatted("However, if you want even sharper shadows (and your hardware can handle the extra")
			imgui.TextUnformatted("work) you can increase shadows resolution in the ASHA settings menu.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("FPS tips")
			imgui.TextUnformatted("--------")
			imgui.TextUnformatted("- If you experience any FPS drop, consider reducing shadows resolution one notch")
			imgui.TextUnformatted("  as it is usually more visually pleasant to keep ASHA running at lower shadows")
			imgui.TextUnformatted("  resolution than using X-Plane default shadows at higher resolutions.")
			imgui.TextUnformatted("- If you only care about cockpit shadows you can also disable scenery shadows in")
			imgui.TextUnformatted("  X-Plane settings menu for an extra FPS boost. But if your hardware can cope with")
			imgui.TextUnformatted("  it, it is highly recommended to keep scenery shadows on, as cities and airports")
			imgui.TextUnformatted("  will look very poor otherwise.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Enjoy ASHA!")
			imgui.TextUnformatted("")
			if shadowres_window or newrelease_window or csmsplit_window or sceneryshadows_window then
				msg = "Next"
			else
				msg = "Understood!"
			end
			if imgui.Button(msg, 100, 30) then
				welcome_window = false
			end
		elseif newrelease_window then
			imgui.TextUnformatted("WALK MODE HAS BEEN INTRODUCED IN 2.2!")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Now you can walk around your aircraft when grounded. Do your walkaround checks")
			imgui.TextUnformatted("from a natural point of view.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("How to activate WALK MODE")
			imgui.TextUnformatted("-------------------------")
			imgui.TextUnformatted("Follow these steps before closing this window so you don't forget!")
			imgui.TextUnformatted("1. Enable the \"allow walk mode\" option in ASHA settings.")
			imgui.TextUnformatted("2. In X-Plane preferences click the Keyboard tab and, in \"Active Profile\" on the")
			imgui.TextUnformatted("   lower left corner select \"ASHA walk mode\". If you can't locate the profile,")
			imgui.TextUnformatted("   try to create and remove another profile to refresh the list or restart X-Plane.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("How to use WALK MODE")
			imgui.TextUnformatted("--------------------")
			imgui.TextUnformatted("After activating walk mode following the steps above, do this:")
			imgui.TextUnformatted("1. Make sure you are grounded and your aircraft has parking brake on. We don't")
			imgui.TextUnformatted("   want any aircraft moving without a pilot ;)")
			imgui.TextUnformatted("2. Press \"Z\" to change to walk mode view.")
			imgui.TextUnformatted("3. Walk around with WASD keys. You can look up/down/sides with the mouse while")
			imgui.TextUnformatted("   holding the right click button.")
			imgui.TextUnformatted("4. Exit walk mode by selecting any other view.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Note about the \"ASHA walk mode\" key profile")
			imgui.TextUnformatted("---------------------------------------------")
			imgui.TextUnformatted("This key profile is generated with your current key assignments. If you already")
			imgui.TextUnformatted("have buttons assignments in W/A/S/D/Z please, change them to other keys and restart")
			imgui.TextUnformatted("X-Plane.")
			imgui.TextUnformatted("")
			if shadowres_window or csmsplit_window or sceneryshadows_window then
				msg = "Next"
			else
				msg = "Understood!"
			end
			if imgui.Button(msg, 100, 30) then
				newrelease_window = false
			end
		elseif shadowres_window then
			imgui.TextUnformatted("ASHA has detected that your shadows resolution is too low.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Current resolution: " .. current_res)
			imgui.TextUnformatted("Recommended resolution: 4096")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("The recommended resolution provides sharper shadows and the best quality vs")
			imgui.TextUnformatted("performance trade off. It might impact performance on low-end computers, but")
			imgui.TextUnformatted("you can reduce it later at ASHA settings menu (at \"Plugins -> FlyWithLua ->")
			imgui.TextUnformatted("-> FlyWithLua Macros -> ASHA settings\") if your FPS drop too much.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Do you want ASHA to set shadows resolution to the recommended value (4096)?")
			imgui.TextUnformatted("")
			-- a click on the yes button calls a function to change shadows resolution
			if imgui.Button("Yes", 100, 30) then
				set_shadow_resolution(4096)
				reboot_required = true
				shadowres_window = false
			end
			imgui.SameLine()
			imgui.SetCursorPosX(150)
			if imgui.Button("No", 100, 30) then
				shadowres_window = false
			end
		elseif csmsplit_window then
			imgui.TextUnformatted("ASHA needs to change a shadow rendering setting in X-Plane in order to properly")
			imgui.TextUnformatted("work. You will always be able to undo this change, since a backup of your current")
			imgui.TextUnformatted("settings will be created. If you decide to remove ASHA in the future, you will have")
			imgui.TextUnformatted("the option to keep this change (which will produce better shadows even with X-Plane")
			imgui.TextUnformatted("defaults) or revert everything to your current status.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("Do you allow ASHA to make this change?")
			imgui.TextUnformatted("")
			if imgui.Button("Yes", 100, 30) then
				set_csmsplits()
				reboot_required = true
				csmsplit_window = false
			end
			imgui.SameLine()
			imgui.SetCursorPosX(150)
			if imgui.Button("No", 100, 30) then
				csmsplit_window = false
			end
		elseif sceneryshadows_window then
			imgui.TextUnformatted("Scenery shadows are disabled!")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("It is higly recommended to enable scenery shadows to enjoy ASHA. Otherwise shadows")
			imgui.TextUnformatted("may look incomplete. The only exceptions where you can leave scenery shadows off")
			imgui.TextUnformatted("are:")
			imgui.TextUnformatted("- You only care about graphics inside the cockpit. Nice graphics in airports and")
			imgui.TextUnformatted("  cities are not important for you.")
			imgui.TextUnformatted("- You fly on a very low-end computer.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("If you want to enable scenery shadows, please tick the option \"Draw shadows on")
			imgui.TextUnformatted("scenery\" in X-Plane's graphic settings menu.")
			imgui.TextUnformatted("")
			if imgui.Button("Understood!", 100, 30) then
				sceneryshadows_window = false
			end
		elseif shadowresmismatch_window then
			local last_setting = tostring(SHADOW_RES)
			if last_setting == "-1" then
				last_setting = "X-Plane defaults"
			end
			local current_setting = tostring(RES_IN_SETTINGS)
			if current_setting == "-1" then
				current_setting = "X-Plane defaults"
			end
			imgui.TextUnformatted("There is a mismatch between your last selected and current shadows resolution:")
			imgui.TextUnformatted("Last selected: " .. last_setting)
			imgui.TextUnformatted("Current: " .. current_setting)
			imgui.TextUnformatted("")
			imgui.TextUnformatted("This could be the consequence of:")
			imgui.TextUnformatted("- X-Plane update restored settings to defaults.")
			imgui.TextUnformatted("- Another plugin interfering ASHA.")
			imgui.TextUnformatted("- You manually edited some X-Plane files.")
			imgui.TextUnformatted("")
			imgui.TextUnformatted("It is recommended to return to ASHA settings, but you can also keep your new")
			imgui.TextUnformatted("resolution.")
			imgui.TextUnformatted("")
			if imgui.Button("Return to ASHA settings (recommended)", 300, 30) then
				set_shadow_resolution(SHADOW_RES)
				reboot_required = true
				shadowresmismatch_window = false
			end
			imgui.SameLine()
			imgui.SetCursorPosX(350)
			if imgui.Button("Keep current resolution", 200, 30) then
				write_asha_settings("shadow_res", RES_IN_SETTINGS)
				shadowresmismatch_window = false
			end
		elseif reboot_required then
			imgui.TextUnformatted("Please, reboot X-Plane for changes to take effect!")
			imgui.TextUnformatted("")
			if imgui.Button("Understood!", 100, 30) then
				reboot_required = false
			end
		else
			float_wnd_destroy(startup_window)
		end
	end
	-- SECOND, FIND OUT WHAT WINDOWS MUST BE SHOWN
	-- trigger for welcome window
	welcome_window = FIRST_RUN
	-- trigger for new release message
	newrelease_window = asha_version_previous ~= asha_version
	-- trigger for shadow resolution change recommendation
	shadowres_window = FIRST_RUN and current_res < 4096
	-- trigger for cascaded shadow maps window
	-- The number of splits has to be changed in the settings.txt file as its associated DataRef is read only. A backup of default X-Plane file is done prior to any change
	-- read settings.txt
	local file = io.open(SYSTEM_DIRECTORY .. "Resources/settings.txt", "r")
	local content = file:read("*a")
	file:close()
	-- read current number of splits
	csmsplit_window = false
	for split in string.gmatch(content, "shadow/csm_split_exterior%s+(%d+)") do
		if tonumber(split) ~= CSM_SPLITS_EXT then
			csmsplit_window = true
			break
		end
	end
	-- trigger for scenery shadows window
	if get("sim/private/controls/shadow/scenery_shadows")==0 and welcome_window then
		sceneryshadows_window = true
	end
	-- trigger for shadow resolution mismatch window
	RES_IN_SETTINGS = res_in_settings_txt()
	if SHADOW_RES ~= RES_IN_SETTINGS then
		shadowresmismatch_window = true
	end
	-- THIRD, POP UP WINDOW
	-- create window
	if welcome_window or newrelease_window or shadowres_window or csmsplit_window or sceneryshadows_window or shadowresmismatch_window then
		startup_window = float_wnd_create(600, 540, 1, true)
		float_wnd_set_position(startup_window, (SCREEN_WIDTH-600)/2, (SCREEN_HIGHT-540)/2)
		float_wnd_set_title(startup_window,  "ASHA")
		float_wnd_set_imgui_builder(startup_window, "draw_startup_window")
	end

	-- FUNCTION: SETTINGS WINDOW
	-- drawing function for settings window
	function draw_settings_window(wnd, x, y)
		-- set a message to remark the current active resolution
		local current1024 = ""
		local current2048 = ""
		local current4096 = ""
		local current6144 = ""
		local current8192 = ""
		local current12288 = ""
		local current16384 = ""
		if current_res == 1024 then
			current1024 = " (your current settings)"
		elseif current_res == 2048 then
			current2048 = " (your current settings)"
		elseif current_res == 4096 then
			current4096 = " (your current settings)"
		elseif current_res == 6144 then
			current6144 = " (your current settings)"
		elseif current_res == 8192 then
			current8192 = " (your current settings)"
		elseif current_res == 12288 then
			current12288 = " (your current settings)"
		elseif current_res == 16384 then
			current16384 = " (your current settings)"
		end
		-- switches
		imgui.TextUnformatted("ASHA features (uncheck to roll back to X-Plane defaults)")
		-- shadows correction on/off switch
		local is_changed, newvalue = imgui.Checkbox("Advanced shadows (cannot be fully disabled, X-Plane defaults will look better now)", SHADOWS_ON)
		if is_changed then
			SHADOWS_ON = newvalue
			write_asha_settings("shadows", newvalue)
		end
		-- jet heat correction on/off switch
		is_changed, newvalue = imgui.Checkbox("Realistic heat effect", HEAT_ON)
		if is_changed then
			HEAT_ON = newvalue
			write_asha_settings("heat", newvalue)
		end
		-- water fft correction on/off switch
		is_changed, newvalue = imgui.Checkbox("Enhanced water simulation (only recommended for 11.50+)", WATER_ON)
		if is_changed then
			WATER_ON = newvalue
			write_asha_settings("water", newvalue)
		end
		-- aurora borealis on/off switch
		is_changed, newvalue = imgui.Checkbox("Aurora borealis (change requires reboot)", AURORA_ON)
		if is_changed then
			AURORA_ON = newvalue
			write_asha_settings("aurora", newvalue)
			-- write new value to X-Plane.prf
			if AURORA_ON then
				xpprf_write("renopt_aurora", 1)
			else
				xpprf_write("renopt_aurora", 0)
			end
		end
		-- fires and balloons on/off switch
		is_changed, newvalue = imgui.Checkbox("Allow forest fires and balloons (change requires reboot)", FIRES_ON)
		if is_changed then
			FIRES_ON = newvalue
			write_asha_settings("fires", newvalue)
			-- write new value to X-Plane.prf
			if FIRES_ON then
				xpprf_write("renopt_fire_ball", 1)
			else
				xpprf_write("renopt_fire_ball", 0)
			end
		end
		-- fps meter on/off switch
		is_changed, newvalue = imgui.Checkbox("Show FPS meter", FPS_METER)
		if is_changed then
			FPS_METER = newvalue
			write_asha_settings("fps_meter", newvalue)
		end
		-- walk mode on/off switch
		is_changed, newvalue = imgui.Checkbox("Allow walk mode (walk with WASD keys, requires \"ASHA walk mode\" key preset active)", WALK_MODE)
		if is_changed then
			WALK_MODE = newvalue
			write_asha_settings("walk_mode", newvalue)
		end
		imgui.TextUnformatted("")
		-- drawing distance
		imgui.TextUnformatted("Shadows drawing distance limit when grounded")
		imgui.TextUnformatted("NOTE: as shadows are never cast farther than the visibility limit, this preference is")
		imgui.TextUnformatted("only active when visibility limit is larger than this setting.")
		is_changed, newvalue = imgui.SliderFloat("", DRAW_DIST, 500, 10000, "%.0f meters", 2)
		if is_changed then
			DRAW_DIST = newvalue
			write_asha_settings("draw_dist", newvalue)
		end
		imgui.TextUnformatted("")
		-- target fps
		imgui.TextUnformatted("In-flight FPS target")
		imgui.TextUnformatted("ASHA actively controls the drawing distance while you fly to match your target FPS or")
		imgui.TextUnformatted("higher. The lower the FPS target, the farther shadows will be cast.")
		is_changed, newvalue = imgui.SliderFloat(" ", TARGET_FPS, 20, 60, "%.0f")
		if is_changed then
			TARGET_FPS = newvalue
			write_asha_settings("target_fps", newvalue)
		end
		imgui.TextUnformatted("")
		-- draw some info text
		imgui.TextUnformatted("Shadow map size (higher size gives sharper shadows, but may impact on FPS)")
		-- radio buttons set the selected resolution and raise a change pending flag when selected
		if imgui.RadioButton("1K" .. current1024, selected_res == 1024) then
			selected_res = 1024
			res_change_pending = 1
		end
		imgui.SameLine()
		imgui.SetCursorPosX(250)
		if imgui.RadioButton("2K" .. current2048, selected_res == 2048) then
			selected_res = 2048
			res_change_pending = 1
		end
		if imgui.RadioButton("4K" .. current4096, selected_res == 4096) then
			selected_res = 4096
			res_change_pending = 1
		end
		imgui.SameLine()
		imgui.SetCursorPosX(250)
		if imgui.RadioButton("6K" .. current6144, selected_res == 6144) then
			selected_res = 6144
			res_change_pending = 1
		end
		if imgui.RadioButton("8K" .. current8192, selected_res == 8192) then
			selected_res = 8192
			res_change_pending = 1
		end
		imgui.SameLine()
		imgui.SetCursorPosX(250)
		if imgui.RadioButton("12K" .. current12288, selected_res == 12288) then
			selected_res = 12288
			res_change_pending = 1
		end
		if imgui.RadioButton("16K" .. current16384, selected_res == 16384) then
			selected_res = 16384
			res_change_pending = 1
		end
		imgui.SameLine()
		imgui.SetCursorPosX(250)
		if imgui.RadioButton("Restore X-Plane defaults", selected_res == -1) then
			selected_res = -1 -- default X-Plane resolution settings flagged as -1 resolution
			res_change_pending = 1
		end
		imgui.TextUnformatted("")
		-- a click on the apply button calls a function to change shadows resolution
		if imgui.Button("Apply resolution change") then
			set_shadow_resolution(selected_res)
			write_asha_settings("shadow_res", selected_res)
			res_change_pending = 2 -- 2 means a change has been applied but X-Plane restart is pending
		end
		imgui.TextUnformatted("")
		if res_change_pending == 1 then
			imgui.TextUnformatted("Press \"Apply resolution change\" to save your selection.")
		elseif res_change_pending == 2 then
			imgui.TextUnformatted("Resolution applied! Please, restart X-Plane for changes to take effect.")
		end
	end
	-- pop up settings window
	function pop_up_settings_window()
		-- create window
		local settings_window = float_wnd_create(650, 600, 1, true)
		float_wnd_set_position(settings_window, (SCREEN_WIDTH-650)/2, (SCREEN_HIGHT-600)/2)
		float_wnd_set_title(settings_window,  "ASHA settings")
		float_wnd_set_imgui_builder(settings_window, "draw_settings_window")
	end

	-- MACRO TO ADD AN ENTRY IN FLYWITHLUA MACROS MENU
	add_macro("ASHA settings", "pop_up_settings_window()")
	
	-- CREATE WALK MODE KEY PRESET
	-- first read current key preset
	local file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane Keys.prf", "r")
	local content = file:read("*a")
	file:close()
	content = string.gsub(content, "\n*%s*$", "")
	-- make necessary key bindings
	if string.match(content, "W <NONE> ")==nil then
		content = content .. "\nW <NONE> sim/general/forward_slow"
	else
		content = string.gsub(content, "W <NONE> [^\n]+", "W <NONE> sim/general/forward_slow")
	end
	if string.match(content, "A <NONE> ")==nil then
		content = content .. "\nA <NONE> sim/general/left_slow"
	else
		content = string.gsub(content, "A <NONE> [^\n]+", "A <NONE> sim/general/left_slow")
	end
	if string.match(content, "S <NONE> ")==nil then
		content = content .. "\nS <NONE> sim/general/backward_slow"
	else
		content = string.gsub(content, "S <NONE> [^\n]+", "S <NONE> sim/general/backward_slow")
	end
	if string.match(content, "D <NONE> ")==nil then
		content = content .. "\nD <NONE> sim/general/right_slow"
	else
		content = string.gsub(content, "D <NONE> [^\n]+", "D <NONE> sim/general/right_slow")
	end
	content = content .. "\n"
	-- create control profiles folder if it does not exist
	if not isdir(SYSTEM_DIRECTORY .. "Output/preferences/control profiles") then
		os.execute("mkdir " .. "\"" .. SYSTEM_DIRECTORY .. "Output/preferences/control profiles\"")
	end
	-- save key preset
	file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/control profiles/ASHA walk mode - Keys.prf", "w")
	file:write(content)
	file:close()
	file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/X-Plane Joystick Settings.prf", "r")
	content = file:read("*a")
	file:close()
	file = io.open(SYSTEM_DIRECTORY .. "Output/preferences/control profiles/ASHA walk mode - Joysticks.prf", "w")
	file:write(content)
	file:close()

	-- FUNCTION: WALK MODE
	local walk_mode_active = false
	local walk_mode_change = os.clock()
	local view_on_ground = false
	local last_key = ""
	local walk_mode_start = -1
	function walk()
		-- wait a delay to free camera, otherwise camera cannot move outside before becoming free
		if walk_mode_start>0 and (os.clock()-walk_mode_start)>0.2 then
			command_once("sim/view/free_camera")
			walk_mode_start = -1
		end
		-- jump iteration if:
		-- - walk mode is disabled
		-- - parking brake is off
		-- - altitude agl > 5 meters
		if not WALK_MODE or parkbrake<1 or altitude_agl>5 then
			return
		end
		-- if "Z" key is pressed, toggle between enabled and disabled walk mode
		last_key = CKEY
		if CKEY=='z' and (os.clock()-walk_mode_change)>0.5 then
			walk_mode_change = os.clock()
			if not walk_mode_active then
				-- walk mode active moves to exterior camera and then frees the camera
				walk_mode_active = true
				set("sim/graphics/view/view_type", 1018)
				walk_mode_start = os.clock()
			else
				walk_mode_active = false
				view_on_ground = false
			end
		end
	end
	-- FUNCTION: SNAP TO GROUND WHILE IN WALK MODE
	local last_y_pos = 999999
	local y_pos
	function snap_to_ground()
		-- jump iteration if:
		-- - not in walk mode
		-- - parking brake is off
		-- - altitude agl > 5 meters
		if not walk_mode_active or not WALK_MODE or parkbrake<1 or altitude_agl>5 then
			return
		end
		-- if camera is no longer in free mode, disable walk mode
		if get("sim/graphics/view/view_type") ~= 1028 then
			walk_mode_active = false
			return
		end
		-- move fast down to ground if view is not on ground, otherwise move slow down to ground
		if not view_on_ground then
			command_once("sim/general/down_fast")
		else
			command_once("sim/general/down_slow")
		end
		-- compensate the down to ground movement if view was already on ground
		y_pos = get("sim/graphics/view/view_y")
		if math.abs(y_pos - last_y_pos) < 0.01 then
			if not view_on_ground then
				command_once("sim/general/up_fast")
				view_on_ground = true
			else
				command_once("sim/general/up_slow")
			end
		end
		-- memorise last camera coordinates
		last_y_pos = y_pos
	end

	-- READ AIRCRAFT GEOMETRY
	-- This is done by inspecting the current .acf file and searching some geometric parameters
	-- read .acf file
	local cockpit_pos = 0
	local cockpit_length = 500
	local wing_span = 40
	local fuselage_length = 40
	local cg_height = 1
	function read_aircraft_geometry()
		local file = io.open(AIRCRAFT_PATH .. AIRCRAFT_FILENAME, "r")
		local acf_content = file:read("*a")
		file:close()
		-- get the relative longitudinal position of cockpit view (acf/_pe_xyz/2 in .acf file, in feet)
		local cockpit_rel_pos = tonumber(string.match(acf_content, "acf/_pe_xyz/2%s+([%d%.-]+)")) / 3.281 -- relative position in meters
		-- get the relative longitudinal position of default CG (acf/_cgZ in .acf file, in feet)
		local CG_rel_pos = tonumber(string.match(acf_content, "acf/_cgZ%s+([%d%.-]+)")) / 3.281 -- relative position in meters
		-- cockpit position relative to default CG (same reference point as camera coordinates)
		cockpit_pos = CG_rel_pos - cockpit_rel_pos
		-- arbitrary cockpit length from cockpit view to cockpit door in meters (used to tell when the active camera is in the cockpit or in the cabin)
		cockpit_length = 3
		-- get all wing elements
		local wing_semilen = {}
		local wing_sweep = {}
		local wing_dihed = {}
		local wing_rootx = {}
		for entry in string.gmatch(acf_content, "P _wing/%d+/_semilen_SEG%s+([-%d%.]+)") do
			wing_semilen[#wing_semilen+1] = tonumber(entry) / 3.281 -- meters
		end
		for entry in string.gmatch(acf_content, "P _wing/%d+/_sweep_design%s+([-%d%.]+)") do
			wing_sweep[#wing_sweep+1] = tonumber(entry) / 180*math.pi -- rad
		end
		for entry in string.gmatch(acf_content, "P _wing/%d+/_dihed_design%s+([-%d%.]+)") do
			wing_dihed[#wing_dihed+1] = tonumber(entry) / 180*math.pi -- rad
		end
		for entry in string.gmatch(acf_content, "P _wing/%d+/_crib_x_arm/0%s+([-%d%.]+)") do
			wing_rootx[#wing_rootx+1] = tonumber(entry) / 3.281 -- meters
		end
		-- calculate span of each wing element and find the maximum
		local semispan = 0
		for idx = 1, math.min(#wing_semilen, #wing_sweep, #wing_dihed, #wing_rootx) do
			local this_span = wing_semilen[idx]*math.cos(wing_sweep[idx])*math.cos(wing_dihed[idx]) + wing_rootx[idx]
			semispan = math.max(semispan, this_span)
		end
		wing_span = semispan * 2 -- meters
		-- get position of start and end of fuselage
		local fus_z_min = 999999
		local fus_z_max = -999999
		for entry in string.gmatch(acf_content, "P _part/%d+/_geo_xyz/%d+,%d+,2%s+([-%d%.]+)") do
			local this_z = tonumber(entry) / 3.281 -- meters
			fus_z_min = math.min(fus_z_min, this_z)
			fus_z_max = math.max(fus_z_max, this_z)
		end
		-- calculate length of fuselage
		fuselage_length = fus_z_max - fus_z_min -- meters
		-- get all landing gears geometry
		local gear_x = {}
		local gear_y = {}
		local gear_z = {}
		local gear_l = {}
		local gear_ang1 = {}
		local gear_ang2 = {}
		local gear_Rtyre = {}
		for entry in string.gmatch(acf_content, "P _gear/%d+/_gear_x%s+([-%d%.]+)") do
			gear_x[#gear_x+1] = tonumber(entry) / 3.281 -- meters
		end
		for entry in string.gmatch(acf_content, "P _gear/%d+/_gear_y%s+([-%d%.]+)") do
			gear_y[#gear_y+1] = tonumber(entry) / 3.281 -- meters
		end
		for entry in string.gmatch(acf_content, "P _gear/%d+/_gear_z%s+([-%d%.]+)") do
			gear_z[#gear_z+1] = tonumber(entry) / 3.281 -- meters
		end
		for entry in string.gmatch(acf_content, "P _gear/%d+/_leg_len%s+([-%d%.]+)") do
			gear_l[#gear_l+1] = tonumber(entry) / 3.281 -- meters
		end
		for entry in string.gmatch(acf_content, "P _gear/%d+/_latE%s+([-%d%.]+)") do
			gear_ang1[#gear_ang1+1] = tonumber(entry) / 180*math.pi -- rad
		end
		for entry in string.gmatch(acf_content, "P _gear/%d+/_lonE%s+([-%d%.]+)") do
			gear_ang2[#gear_ang2+1] = tonumber(entry) / 180*math.pi -- rad
		end
		for entry in string.gmatch(acf_content, "P _gear/%d+/_tire_radius%s+([-%d%.]+)") do
			gear_Rtyre[#gear_Rtyre+1] = tonumber(entry) / 3.281 -- meters
		end
		-- find landing gear closer to the CG
		local N = 1
		local min_CGdist = 999999
		for idx = 1, #gear_z do
			local CGdist = math.abs(gear_z[idx] - CG_rel_pos)
			if (CGdist < min_CGdist) and (gear_l[idx] > 0) then
				N = idx
			end
		end
		-- calculate CG height at the landing gear closer to CG
		local leg_y = gear_l[N]*math.cos(gear_ang1[N])*math.cos(gear_ang2[N])
		local full_y = gear_y[N] - leg_y - gear_Rtyre[N]
		local CG_y = tonumber(string.match(acf_content, "acf/_cgY%s+([%d%.-]+)")) / 3.281 -- relative position in meters
		cg_height = CG_y - full_y
	end
	local success, exception = pcall(read_aircraft_geometry)
	if not success then
		logMsg(os.date("%c") .. " ASHA: can't read aircraft geometry")
		logMsg(os.date("%c") .. " ASHA: " .. exception)
		logMsg(os.date("%c") .. " ASHA: please, report this issue to ASHA developer (Alreiber) via PM in x-plane.org attaching a copy of " .. AIRCRAFT_PATH .. AIRCRAFT_FILENAME)
	end

	-- FUNCTION: RESTORE ALL SHADOW DEFAULTS (user can disable ASHA from its menu)
	local is_shadow_defaults = SHADOWS_ON
	function restore_default_shadows()
		-- skip restore if X-Plane already on defaults
		if is_shadow_defaults then
			return
		end
		-- as we know renopt_scenery_shadows and renopt_effects_04 (obtained at the top of this script), apply default X-Plane settings
		local def_far_limit_exterior = 0
		local def_far_limit_interior = 0
		local def_far_limit = 0
		local def_fade_ratio = 0.7
		if renopt_scenery_shadows == 0 then
			def_far_limit_exterior = 500
			def_far_limit_interior = 500
			def_far_limit = 500
			def_fade_ratio = 0.999
		elseif renopt_effects_04 == 0 then
			def_far_limit_exterior = 120
			def_far_limit_interior = 120
			def_far_limit = 750
		elseif renopt_effects_04 == 1 then
			def_far_limit_exterior = 250
			def_far_limit_interior = 250
			def_far_limit = 750
		elseif renopt_effects_04 >= 2 then
			def_far_limit_exterior = 500
			def_far_limit_interior = 500
			def_far_limit = 1500
		end
		set("sim/private/controls/shadow/cockpit_near_adjust", 1)
		set("sim/private/controls/skyc/max_shadow_angle", -89)
		set("sim/private/controls/skyc/min_shadow_angle", -90)
		set("sim/private/controls/shadow/total_fade_ratio", def_fade_ratio)
		set("sim/private/controls/shadow/csm/lambda", 1)
		if is1150 then
			set("sim/private/controls/shadow/csm/far_limit_exterior", def_far_limit_exterior)
			set("sim/private/controls/shadow/csm/far_limit_interior", def_far_limit_interior)
			set("sim/private/controls/shadow/csm/near_limit_exterior", 4)
			set("sim/private/controls/shadow/csm/near_limit_interior", 2)
		else
			set("sim/private/controls/shadow/csm/far_limit", def_far_limit)
		end
	end
	
	-- FUNCTION: DRAWING DISTANCE CONTROLLER
	local Pgain = 100
	local Igain = 500
	local Ierr = 0
	local Perr = 0
	local PI_far = 5000
	local t0 = os.clock()
	local tdelta = 0
	local far_max = 10000
	-- PI controller matching FPS target
	function drawdist_controller()
		-- calculate proportional and integral errors
		Perr = fps - TARGET_FPS
		tdelta = os.clock() - t0
		t0 = os.clock()
		Ierr = Ierr + Perr*tdelta
		-- calculate PI output
		PI_far = Pgain*Perr + Igain*Ierr
		-- limit output to 100 times the altitude (empirical, to avoid a noticeably blurry third split)
		far_max = math.min(math.max(20*altitude_agl, 500), 250000)
		if (PI_far > far_max) or (PI_far < 500) or (gs < 25) then
			Ierr = Ierr - Perr*tdelta
			PI_far = math.max(math.min(Pgain*Perr + Igain*Ierr, far_max), 500)
		end
	end
	
	-- FUNCTION: CALCULATE AND APPLY SHADOWS CORRECTION
	-- this function is executed regularly to recalculate corrections based on active camera location
	local cam_x = 0
	local cam_y = 0
	local cam_z = 0
	local cam_dist = 0
	local cam_agl = 0
	local is_in_cockpit = 0
	local is_look_side = 0
	local DvisGeom = 0
	local DvisAccu = 0
	local Dvis = 0
	local D0sun = 0
	local D0close = 0
	local D0low = 0
	local Dzlow = 0
	local Dxlow = 0
	local Clog = 0
	local Cuni = 0
	local Ci = 0
	local shadows_correction_last_exec = 0
	local shadows_correction_period = 0.25 -- seconds
	function shadows_correction()
		-- skip all corrections and set default X-Plane settings if Advanced Shadows are turned off
		if not SHADOWS_ON then
			restore_default_shadows()
			is_shadow_defaults = true
			return
		end
		is_shadow_defaults = false
		-- skip all corrections if last execution was too recent (one execution each "shadows_correction_period" seconds)
		if os.clock() < shadows_correction_last_exec + shadows_correction_period then
			return
		end
		shadows_correction_last_exec = os.clock()

		-- CAMERA POSITION
		-- calculate active camera position relative to aircraft default CG, in meters
		cam_x = camera_x - local_x
		cam_y = camera_y - local_y
		cam_z = camera_z - local_z
		-- total distance between active camera and aircraft default CG
		cam_dist = math.sqrt(cam_x^2 + cam_y^2 + cam_z^2)
		-- active camera height above ground level
		cam_agl = math.max(cam_y + altitude_agl + cg_height, 0.01)
		
		-- CAMERA LOCATION INSIDE AIRCRAFT
		-- this flag is 1 if the active camera is inside the cockpit, or 0 if it is in the cabin or external to the aircraft
		is_in_cockpit = 0
		if is_external_view == 0 and -int_camera_z > cockpit_pos-cockpit_length then
			is_in_cockpit = 1
		end
		-- check if camera inside cabin is looking outside (sides) or inside
		is_look_side = 0
		if (int_camera_heading > 20 and int_camera_heading < 160 and int_camera_x > 0.5) or (int_camera_heading > 200 and int_camera_heading < 340 and int_camera_x < -0.5) then
			is_look_side = 1
		end
		
		-- CALCULATE CASCADED SHADOW MAPS
		-- cam_height is determined by agl if aircraft is grounded or above aircraft's CG if airborne
		if altitude_agl < 20 then
			cam_h = cam_agl
		else
			cam_h = math.abs(cam_y)
			cam_pitch = -math.abs(cam_pitch)
		end
		-- fixed parameters
		local near = 0.84 -- near plane
		local splits = csm_split_exterior -- number of splits
		-- screen resolution (limited to 1080 to avoid excessive shadow rendering resolution)
		local screen_h = math.min(SCREEN_HIGHT, 1080)
		local screen_w = screen_h/SCREEN_HIGHT*SCREEN_WIDTH
		-- max drawing distance (visibility limit, defined as distance where dInf meters in depth are less than 1 vertical pixel on screen)
		DvisGeom = cam_agl / math.sin(math.max(- cam_pitch - vert_fov/2, 0)/180*math.pi) / math.cos(horz_fov/2/180*math.pi) -- meters (to inf if horizon is visible)
		local dInf = 50 -- meters
		local deltaInf = vert_fov/screen_h -- degrees (for 1 vertical pixels on screen)
		DvisAccu = math.max((- dInf + math.sqrt(math.max(dInf^2 - 4*(cam_agl^2 - dInf*cam_agl/math.tan(deltaInf/180*math.pi)), 0))) / 2, 0) -- meters (clipped to the ground)
		DvisAccu = math.sqrt(DvisAccu^2 + cam_agl^2) -- meters (distance to the camera)
		Dvis = math.min(DvisAccu + 500, DvisGeom) -- actual visibility is minimum between geometric visibility and accuracy-limited visibility
		-- far plane
		-- on ground (<25 m agl and <25 m/s): set to the actual visibility limit to a maximum of DRAW_DIST meters
		-- flying: drawing distance controller takes over
		local far = PI_far -- meters
		if altitude_agl < 20 and gs < 25 then
			local far2 = math.min(math.max(math.min(Dvis*1.25, Dvis+50), 100), DRAW_DIST) -- meters
			local k = math.max(math.min(gs/25, 1), 0)
			far = k*far + (1-k)*far2
		end
		-- ADJUST FIRST SPLIT
		local split_no = 1 -- this split number
		-- drawing distance that makes the nearest chunk of shadow to be rendered with one shadow pixel in one screen pixel
		local shadow_pixels = 2 -- number of shadow pixels that fall into one screen pixel
		D0sun = current_res/shadow_pixels * (cam_h/math.tan((- cam_pitch + vert_fov/2 - vert_fov/screen_h)/180*math.pi) - cam_h/math.tan((- cam_pitch + vert_fov/2)/180*math.pi)) -- meters
		local hbuild = 20 -- average height of surrounding buildings
		D0close = math.sqrt(D0sun^2 + hbuild^2/math.tan(math.max(sun_pitch, 0)/180*math.pi)^2 - 2*math.sqrt(D0sun^2 - cam_h^2)*hbuild/math.tan(math.max(sun_pitch, 0)/180*math.pi)) -- meters
		local rho = 0.65
		D0low = rho*D0sun + (1-rho)*D0close -- meters
		-- drawing distance where "dz" meters on ground (z-axis, depth) are rendered to "deltaz" degrees of vertical FOV
		local dz = far/current_res -- meters (per one pixel)
		local deltaz = vert_fov/screen_h*1 -- degrees (for 1 vertical pixels on screen)
		Dzlow = math.max((- dz + math.sqrt(math.max(dz^2 - 4*(cam_h^2 - dz*cam_h/math.tan(deltaz/180*math.pi)), 0))) / 2, 0) -- meters (clipped to the ground)
		Dzlow = math.sqrt(Dzlow^2 + cam_h^2) -- meters (distance to the camera)
		-- drawing distance where "d" meters on ground (x-axis, width) are rendered to "delta" degrees of horizontal FOV
		local dx = math.tan(horz_fov/2/180*math.pi)*far/current_res -- meters (per one pixel)
		local deltax = horz_fov/screen_w*1 -- degree (for 1 horizontal pixels on screen)
		Dxlow = dx / (math.tan(horz_fov/2/180*math.pi) - math.tan((horz_fov/2 - deltax)/180*math.pi)) -- meters (clipped to the ground)
		Dxlow = math.sqrt(Dxlow^2 + cam_h^2) -- meters (distance to the camera)
		-- desired distance for this split
		local Dlow = math.min(math.max(Dzlow, Dxlow), D0low)
		-- logarithmic and uniform splits distance
		Clog = near*(far/near)^(split_no/splits) -- log split distance
		Cuni = near+(far-near)*split_no/splits -- uniform split distance
		-- set a minimum and max split distance
		local Cimin = 5 + math.max(wing_span, fuselage_length) -- meters (max aircraft dimension with a 10 meter margin)
		local Cimax = 999999 -- meters
		if cam_dist>5000 then -- if distance to aircraft is greater than 5000 assume the camera is not pointing to the aircraft
			Cimax = 75
		end
		-- limit split with minimum and maximum split distance
		Ci = math.min(math.max(Dlow, Cimin), Cimax) -- meters
		-- limit this split distance between logarithmic and uniform (only uniform limit active)
		-- Ci = math.max(math.min(Ci, Cuni), Clog)
		Ci = math.min(Ci, Cuni)
		-- calculate lambda
		local lambda = (Ci - Cuni) / (Clog - Cuni)
		
		-- SET LAMBDA FOR INTERIOR VIEWS (lambda equivalent to second split as first split is placed manually with the code below)
		if is_external_view==0 then
			if altitude_agl < 25 then
				Cimin = 5
				Cimax = math.max(math.min(10 + (cam_agl-1.5)/2.5*10, 30), 5)
			else
				Cimin = 25
				Cimax = 200
			end
			split_no = 2
			local Clog2 = near*(far/near)^(split_no/splits)
			local Cuni2 = near+(far-near)*split_no/splits
			Ci = lambda*(Clog2 - Cuni2) + Cuni2
			Ci = math.max(math.min(Ci, Cimax), Cimin)
			lambda = (Ci - Cuni) / (Clog - Cuni)
		end
		
		-- SET NEAR MAP ADJUSTMENT
		local adjust = 1 -- in exterior view
		if is_in_cockpit==1 then -- in cockpit
			adjust = 1.7
		elseif is_external_view==0 and is_look_side==1 then -- in cabin, looking to a side
			adjust = 8
		elseif is_external_view==0 then -- in cabin, looking inside
			adjust = 1
			lambda = 1
			far = fuselage_length
		end
		
		-- NEAR LIMIT ADJUSTMENT
		local nearext = 4
		local nearint = 2
		if is_external_view==1 then
			nearext = 0.5
			nearint = 0.5
		end

		-- APPLY CORRECTIONS CALCULATED ABOVE
		set("sim/private/controls/shadow/cockpit_near_adjust", adjust)
		set("sim/private/controls/shadow/csm/lambda", lambda)
		if is1150 then
			set("sim/private/controls/shadow/csm/far_limit_exterior", far)
			set("sim/private/controls/shadow/csm/far_limit_interior", far)
			set("sim/private/controls/shadow/csm/near_limit_exterior", nearext)
			set("sim/private/controls/shadow/csm/near_limit_interior", nearint)
		else
			set("sim/private/controls/shadow/csm/far_limit", far)
		end
		-- CHANGE MIN/MAX SUN ANGLE TO RENDER SHADOWS
		-- By default shadows are not rendered at low Sun angles. This covers from sunrise to sunset
		local max_angle = -3
		local min_angle = -3
		if not is1150 then
			max_angle = -89
			min_angle = -90
		end
		set("sim/private/controls/skyc/max_shadow_angle", max_angle)
		set("sim/private/controls/skyc/min_shadow_angle", min_angle)
	end
	
	-- FUNCTION: CHANGE SHADOW FADE RATIO
	-- Farthest shadow map is faded starting at this ratio to provide a smooth transition between shadow and no-shadow objects/scenery
	-- (e.g. if ratio is the default 0.7 and farthest shadow map spans 1000 meters, the fade will start at 700 meters and end at 1000)
	function set_fade_ratio()
		local fade = 0.6
		if not SHADOWS_ON then
			fade = 0.7
		end
		set("sim/private/controls/shadow/total_fade_ratio", fade)
	end
	
	-- FUNCTION: AUXILIAR FUNCTION TO DO LINEAR INTERPOLATIONS
	local function interp(x, y, xi)
		-- find interval in x where xi is contained
		local interval_right = 1
		while x[interval_right] <= xi do
			interval_right = interval_right + 1
		end
		local my_x = {x[interval_right-1], x[interval_right]}
		local my_y = {y[interval_right-1], y[interval_right]}
		-- interpolate linearly within the interval
		local w = (xi - my_x[1]) / (my_x[2] - my_x[1])
		local yi = (my_y[2] - my_y[1]) * w + my_y[1]
		return yi
	end

	-- FUNCTION: DO ADDITIONAL TWEAKS
	local is_heat_default = HEAT_ON
	local is_aurora_default = AURORA_ON
	local is_fires_default = FIRES_ON
	local is_water_default = WATER_ON
	local asha_tweaks_last_exec = 0
	local asha_tweaks_period = 0.25 -- seconds
	function asha_tweaks()
		-- skip all tweaks if last execution was too recent (one execution each "asha_tweaks_period" seconds)
		if os.clock() < asha_tweaks_last_exec + asha_tweaks_period then
			return
		end
		asha_tweaks_last_exec = os.clock()
		
		-- JET HEAT REDUCED OPACITY
		local normal_blur = 2
		local zoom_blur = 6
		if HEAT_ON then
			normal_blur = 0
			zoom_blur = 0
		end
		if HEAT_ON or not is_heat_default then
			set("sim/private/controls/heat/normal_blur", normal_blur)
			set("sim/private/controls/heat/zoom_blur", zoom_blur)
		end
		if HEAT_ON then
			is_heat_default = false
		else
			is_heat_default = true
		end
	end
	
	-- FUNCTION: DO WATER FFT ENHANCEMENTS
	local amp1 = 1
	local amp2 = 1
	local amp3 = 1
	local amp4 = 1
	local sca1 = 1
	local sca2 = 8
	local sca3 = 64
	local sca4 = 256
	local noise_speed = 9.81
	function asha_water()
		if WATER_ON then
			-- water fft parameters
			amp3 = 1
			sca3 = 200
			sca4 = 1000
			-- water speed
			noise_speed = 2
		end
		if WATER_ON or not is_water_default then
			set("sim/private/controls/water/fft_amp3", amp1)
			set("sim/private/controls/water/fft_amp3", amp2)
			set("sim/private/controls/water/fft_amp3", amp3)
			set("sim/private/controls/water/fft_amp3", amp4)
			set("sim/private/controls/water/fft_scale3", sca1)
			set("sim/private/controls/water/fft_scale3", sca2)
			set("sim/private/controls/water/fft_scale3", sca3)
			set("sim/private/controls/water/fft_scale4", sca4)
			set("sim/private/controls/water/noise_speed", noise_speed)
		end
		if WATER_ON then
			is_water_default = false
		else
			is_water_default = true
		end
	end

	-- FUNCTION: DRAW DEBUG MODE OVERLAY
	function draw_debug()
		-- coordinates from the bottom left corner of the overlay box in pixels
		local x = 100
		local y = SCREEN_HIGHT-100
		-- draw the overlay
		glColor4f(0.3, 1, 0.3, 1)
		draw_string_Helvetica_18(x, y, "Advanced Shadows " .. asha_version .. " - " .. asha_build .. " build")
		graphics.draw_line(x, y-10, x+450, y-10)
		if SHADOWS_ON then
			draw_string_Helvetica_18(x, y-40, "Status: ON ")
		else
			draw_string_Helvetica_18(x, y-40, "Status: OFF ")
		end
		draw_string_Helvetica_18(x, y-60, "Camera position")
		draw_string_Helvetica_18(x, y-80, "(in meters)")
		draw_string_Helvetica_18(x+160, y-60, string.format("x_from_CG = %.1f", cam_x))
		draw_string_Helvetica_18(x+160, y-80, string.format("y_from_CG = %.1f", cam_y))
		draw_string_Helvetica_18(x+160, y-100, string.format("z_from_CG = %.1f", cam_z))
		draw_string_Helvetica_18(x+160, y-120, string.format("dist_to_aircraft_CG = %.1f", cam_dist))
		draw_string_Helvetica_18(x+160, y-140, string.format("agl = %.1f", cam_agl))
		draw_string_Helvetica_18(x, y-170, "Internal data")
		draw_string_Helvetica_18(x+130, y-170, string.format("fuselage_length [m] = %.1f", fuselage_length))
		draw_string_Helvetica_18(x+130, y-190, string.format("wing_span [m] = %.1f", wing_span))
		draw_string_Helvetica_18(x+130, y-210, string.format("CG_height [m] = %.1f", cg_height))
		draw_string_Helvetica_18(x+130, y-230, string.format("ac_altitude_agl [m] = %.1f", altitude_agl))
		draw_string_Helvetica_18(x+130, y-250, string.format("is_external_view = %.0f", is_external_view))
		draw_string_Helvetica_18(x+130, y-270, string.format("is_in_cockpit = %.0f", is_in_cockpit))
		draw_string_Helvetica_18(x+130, y-290, string.format("is_look_side = %.0f", is_look_side))
		draw_string_Helvetica_18(x+130, y-310, string.format("shadow_cam_size = %.0f", current_res))
		draw_string_Helvetica_18(x+130, y-330, string.format("cam_pitch = %.1f", -cam_pitch))
		draw_string_Helvetica_18(x+130, y-350, string.format("vert_fov = %.1f", vert_fov))
		draw_string_Helvetica_18(x+130, y-370, string.format("horz_fov = %.1f", horz_fov))
		draw_string_Helvetica_18(x+130, y-390, string.format("csm_split_exterior = %.0f", csm_split_exterior))
		draw_string_Helvetica_18(x+130, y-410, string.format("visibility_geometric = %.1f", DvisGeom))
		draw_string_Helvetica_18(x+130, y-430, string.format("visibility_accuracy = %.1f", DvisAccu))
		draw_string_Helvetica_18(x+130, y-450, string.format("visibility_actual = %.1f", Dvis))
		draw_string_Helvetica_18(x+130, y-470, string.format("D0sun = %.1f", D0sun))
		draw_string_Helvetica_18(x+130, y-490, string.format("D0close = %.1f", D0close))
		draw_string_Helvetica_18(x+130, y-510, string.format("D0low = %.1f", D0low))
		draw_string_Helvetica_18(x+130, y-530, string.format("Dzlow = %.1f", Dzlow))
		draw_string_Helvetica_18(x+130, y-550, string.format("Dxlow = %.1f", Dxlow))
		draw_string_Helvetica_18(x+130, y-570, string.format("Clog = %.2f", Clog))
		draw_string_Helvetica_18(x+130, y-590, string.format("Cuni = %.2f", Cuni))
		draw_string_Helvetica_18(x+130, y-610, string.format("Ci = %.2f", Ci))
		draw_string_Helvetica_18(x+130, y-630, string.format("walk_mode_active = " .. tostring(walk_mode_active)))
		draw_string_Helvetica_18(x, y-660, "Drawing distance controller")
		draw_string_Helvetica_18(x+240, y-660, string.format("TARGET_FPS = %.0f", TARGET_FPS))
		draw_string_Helvetica_18(x+240, y-680, string.format("Perr = %.2f", Perr))
		draw_string_Helvetica_18(x+240, y-700, string.format("Ierr = %.2f", Ierr))
		draw_string_Helvetica_18(x+240, y-720, string.format("PI_far = %.0f", PI_far))
		draw_string_Helvetica_18(x, y-750, "Applied corrections")
		draw_string_Helvetica_18(x+190, y-750, string.format("cockpit_near_adjust = %.1f", cockpit_near_adjust))
		draw_string_Helvetica_18(x+190, y-770, string.format("csm_lambda = %.3f", csm_lambda))
		if is1150 then
			draw_string_Helvetica_18(x+190, y-790, string.format("far_limit_exterior = %.0f", far_limit_exterior))
			draw_string_Helvetica_18(x+190, y-810, string.format("far_limit_interior = %.0f", far_limit_interior))
		else
			draw_string_Helvetica_18(x+190, y-790, string.format("far_limit = %.0f", far_limit))
		end
	end

	-- FUNCTION: SHOW FPS METER
	local counter_x = 0
	local counter_y = 0
	function show_fps()
		if not FPS_METER then
			return
		end
		counter_x = SCREEN_WIDTH - 60
		counter_y = SCREEN_HIGHT - 40--15
		graphics.set_color(0, 0, 0, 0.5)
		graphics.draw_rectangle(counter_x-10, counter_y-10, SCREEN_WIDTH, SCREEN_HIGHT)
		if fps <= 19 then
			draw_string(counter_x, counter_y, string.format("FPS: %.0f", fps), "red")
		elseif fps <= 30 then
			draw_string(counter_x, counter_y, string.format("FPS: %.0f", fps), "yellow")
		else
			draw_string(counter_x, counter_y, string.format("FPS: %.0f", fps), "green")
		end
	end

	-- EXECUTE TWEAKS REGULARLY (triggered every frame, but internally triggered at a slower pace)
	-- shadows correction
	do_every_frame("shadows_correction()")
	-- other tweaks
	do_every_frame("asha_tweaks()")
	-- water enhancement
	do_every_frame("asha_water()")
	-- fade ratio in a separate thread run every frame due to interference with xEnviro
	do_every_frame("set_fade_ratio()")
	-- fps calculation
	do_every_frame("count_frames()")
	-- display fps meter
	do_every_draw("show_fps()")
	-- drawing distance controller
	do_every_frame("drawdist_controller()")
	-- walk mode
	do_on_keystroke("walk()")
	-- snap to ground while in walk mode
	do_every_frame("snap_to_ground()")
	-- execute debug mode every frame only if required
	if DEBUG_MODE then
		do_every_draw("draw_debug()")
	end
end

-- CALL MAIN LOOP (will be executed just once)
do_often("asha_main_loop()")